import math
import numpy as np


class AD:
    """
    This class is used to represent functions (and their variables) for automatic differentiation.
    In order to get the derivative of a function with respect to a specific variable, that variable must be initialized as an AD object before performing any operation in which this variable is involved.

    Vector functions are also initialized as an AD object.

    ...

    Parameters
    ==========
    value : int, float, np.ndarray or list
        Values or functions used to construct a new variable or function object:
          - if value is int or float or np.ndarray         --> variable or scalar function
          - if value is list(int, float, AD or np.ndarray) --> vector function

    der : int, float, np.ndarray or list, optional
        Derivatives with respect to each of the variables of a function (default is 0).
        This parameter is only needed when instantiating variables and not for functions.

    name : str, optional
        Names of the variables (default is 'x').
        This parameter is only needed when instantiating variables and not for functions.

    Attributes
    ==========
    val : np.ndarray (scalar functions) or list (vector functions)
        Value(s) of the variable(s) where the functions are evaluated.
    der : np.ndarray (scalar functions) or list (vector functions)
        Jacobian: Derivative(s) with respect to each of the variable(s)
    name : list of strings
        Name(s) of the variable(s)

    Examples
    ==========
    # Scalar input (x)
    >>> x = AD(2,1,'x')
    >>> f = 7*x + 0.3
    >>> f
    Numerical Value is:
    [[14.3]],
    Jacobian is:
    [[7.]],
    Name is:
    ['x']

    # Vector input (x,y)
    >>> x = AD(2,1,'x')
    >>> y = AD(3,1,'y')
    >>> f = 5*x + 4*y + 0.5
    >>> f
    Numerical Value is:
    [[22.5]],
    Jacobian is:
    [[5. 4.]],
    Name is:
    ['x', 'y']

    # Vector input (x,y) and Vector output (f1,f2,f3)
    >>> x = AD(2,1,'x')
    >>> y = AD(3,1,'y')
    >>> f = AD([5*x+4*y+0.5, 43*x, 7]) #f1,f2,f3 = 5*x+4*y+0.5, 43*x, 7
    >>> f
    Numerical Value is:
    [22.5, 86.0, 7],
    Jacobian is:
    [[5.0, 4.0], [43.0, 0], [0, 0]],
    Name is:
    ['x', 'y']


    Methods
    ==========
    # AD object-related methods
    __init__(self, value, der=0, name='x'):  Constructs the necessary attributes of an AD object representing a variable or a function.
    __repr__(self):  Return the canonical string representation of the object.
    sort(self, order):  Sort the derivatives and variable names of an AD object by the specified order.

    # Basic operations
    __add__(self, other):  Perform addition on an AD object.
    __radd__(self, other):  Perform reverse addition on an AD object.
    __neg__(self):  Perform negation on AD objects.
    __sub__(self, other):  Perform subtraction on an AD object.
    __rsub__(self, other):  Perform reverse subtraction on an AD object.
    __mul__(self, other):  Perform multiplication on an AD object.
    __rmul__(self, other):  Perform reverse multiplication on an AD object.
    __truediv__(self, other):  Perform true division on an AD object.
    __rtruediv__(self, other):  Perform reverse true division on an AD object.
    __pow__(self, n):  Raise an AD object to the power of n.
    __rpow__(self, n):  Raise a number to the power of an AD object.

    # Comparisons
    __lt__(self, other):  Perform "less than" comparison on an AD object.
    __gt__(self, other):  Perform "greater omparison on an AD object.
    __le__(self, other):  Perform "less or equal than" comparison on an AD object.
    __ge__(self, other):  Perform "greater or equal than" comparison on an AD object.
    __eq__(self, other):  Perform "equality" comparison on an AD object.
    __ne__(self, other):  Perform "inequality" comparison on an AD object.

    # Elementary functions
    tan(self):  Compute the tangent of an AD object.
    sin(self):  Compute the sine of an AD object.
    cos(self):  Compute the cosine of an AD object.
    exp(self):  Compute the exponential of an AD object.
    ln(self):  Compute the natural logarithm of an AD object.
    ln_base(self):  Compute the base-specific logarithm of an AD object.
    sinh(self):  Compute the hyperbolic sine of an AD object.
    cosh(self):  Compute the hyperbolic cosine of an AD object.
    tanh(self):  Compute the hyperbolic tangent of an AD object.
    arcsin(self):  Compute the arcsine (inverse of sine) of an AD object.
    arccos(self):  Compute the arccosine (inverse of cosine) of an AD object.
    arctan(self):  Compute the arctangent (inverse of tangent) of an AD object.
    logistic(self):   Apply the sigmoid function to an AD object, defined as: sigmoid(x) =  1/(1+e**(-x))
    sqrt(self): Compute the square root of an AD object.
    """

    def __init__(self, value, der=0, name='x'):
        """
        Constructs the necessary attributes of an AD object representing a variable or a function.

        Parameters
        ----------
        value : int, float, np.ndarray or list
            Values or functions used to construct a new variable or function object
              - if value is int or float or np.ndarray         --> scalar function
              - if value is list(int, float, AD or np.ndarray) --> vector function

        der : int, float, np.ndarray or list, optional
            Derivatives with respect to each of the variables of a function (default is 0).
            This parameter is only needed when instantiating variables and not for functions.

        name : str, optional
            Names of the variables (default is 'x').
            This parameter is only needed when instantiating variables and not for functions.

        Returns
        -------
        AD object representing a variable or a function, with the corresponding derivatives and variable names.

        Examples
        --------
        # Scalar input (x)
        >>> x = AD(2,1,'x')
        >>> x
        Numerical Value is:
        [[2.]],
        Jacobian is:
        [[1.]],
        Name is:
        ['x']
        # Vector input (x,y)
        >>> x = AD(2,1,'x')
        >>> y = AD(3,1,'y')
        >>> f = 5*x + 4*y + 0.5
        >>> f
        Numerical Value is:
        [[22.5]],
        Jacobian is:
        [[5. 4.]],
        Name is:
        ['x', 'y']
        # Vector input (x,y) and Vector output (f1,f2,f3)
        >>> x = AD(2,1,'x')
        >>> y = AD(3,1,'y')
        >>> f = AD([5*x+4*y+0.5, 43*x, 7]) #f1,f2,f3 = 5*x+4*y+0.5, 43*x, 7
        >>> f
        Numerical Value is:
        [22.5, 86.0, 7],
        Jacobian is:
        [[5.0, 4.0], [43.0, 0], [0, 0]],
        Name is:
        ['x', 'y']
        """
        self.val = None
        self.der = None
        self.name = None

        if isinstance(value, list):
            names = []
            for AD_function in value:
                try:
                    names.append(AD_function.name)
                except AttributeError:
                    continue  # if just list names never has anything appended
            # unique_names = set(np.asarray(names).flatten())  # reference
            try:
                unique_names = list(set(np.array(np.concatenate(names, axis=0))))  # reference
            except ValueError:
                unique_names = set(np.asarray(names).flatten())
            global_value = []  # vector of values
            global_jacobian = []  # matrix of derivatives, every derivative of the list should be one row
            for AD_function in value:
                try:
                    # print('The value we are appending is ', AD_function.val)
                    global_value.append(AD_function.val[0][0])
                except AttributeError:
                    global_value.append(AD_function)  # constant
            for AD_function in value:
                AD_jacobian = []
                for var in unique_names:
                    try:
                        # print('We are looking for ', var)
                        # print('Inside', AD_function.name)
                        index = AD_function.name.index(var)  # ['x', 'y'] vs ['y', 'x']
                        # print('It is ', index)
                        # print('The derivative of the AD var is ', AD_function.der)
                        AD_jacobian.append(AD_function.der[0][index])
                    except:
                        AD_jacobian.append(0)
                    # print('...', AD_jacobian)
                global_jacobian.append(AD_jacobian)


            # if len(unique_names) != 0:
            self.val = np.array(global_value).reshape(-1, 1)
            self.der = np.array(global_jacobian)
            self.name = unique_names
            # else:
            # self.val = np.array(value).reshape(len(value), 1)

        if self.val is None:
            if isinstance(value, float) or isinstance(value, int):
                value = np.array(value, dtype=np.float64).reshape(1, -1)
            elif isinstance(value, np.ndarray):
                value = value.reshape(value.shape[0], 1)
            else:
                raise TypeError(f'Value should be int or float and not {type(value)}')

            self.val = np.array(value, dtype=np.float64)

        if self.der is None:
            if isinstance(der, float) or isinstance(der, int):
                der = np.array(der, dtype=np.float64).reshape(1, -1)
            elif isinstance(der, list) and len(der) == 1:
                der = np.array(der).reshape(len(der), 1)
            elif isinstance(der, np.ndarray) and der.shape[0] == 1:
                der = der.reshape(der.shape[0], -1)
            elif isinstance(der, np.ndarray) and len(der.shape) == 2:
                der = der
            else:
                raise TypeError(f'Derivative should be int or float and not {type(der)}')

            self.der = np.array(der, dtype=np.float64)

        if self.name is None:
            if isinstance(name, list):
                self.name = name
            else:
                self.name = [name]

    def __repr__(self):
        """
        Get a string representation of the AD object. It prints:
           - the values where the function is evaluated (self.val)
           - the Jacobian (self.der)
           - the names of the function variables (self.name)

        Returns
        -------
        String representing the AD object's values, Jacobian and variable names.

        Example
        -------
        >>> x = AD(2,1,'x')
        >>> y = AD(3,1,'y')
        >>> f = AD([5*x+4*y+0.5, 11*x*y])
        >>> f
        Numerical Value is:
        [22.5, 66.0],
        Jacobian is:
        [[5.0, 4.0], [33.0, 22.0]],
        Name is:
        ['x', 'y']
        """
        val = self.val
        der = self.der
        name = self.name
        return 'Numerical Value is:\n{}, \nJacobian is:\n{}, \nName is:\n{}'.format(val, der, name)

    def sort(self, order):
        """
        Sort the derivatives and variable names of an AD object by the specified order.

        Parameters
        ----------
        order : list(str)
            Names of the variables in the desired order.

        Returns
        -------
        AD object with self.val, self.der and self.name in the desired order.

        Example
        -------
        >>> x = AD(2,1,'x')
        >>> y = AD(3,1,'y')
        >>> z = AD(4,1,'z')
        >>> f = AD([5*x+4*y+3*z, x*y*z])
        >>> print(f)
        >>> f.sort(['x', 'y', 'z'])
        >>> print(f)
        Numerical Value is:
        [34.0, 24.0],
        Jacobian is:
        [[5.0, 3.0, 4.0], [12.0, 6.0, 8.0]],
        Name is:
        ['x', 'z', 'y']
        Numerical Value is:
        [34.0, 24.0],
        Jacobian is:
        [[5.0, 4.0, 3.0], [12.0, 8.0, 6.0]],
        Name is:
        ['x', 'y', 'z']
        """
        if not isinstance(order, list) and not isinstance(order, np.ndarray):
            raise TypeError('Order should be an array-like composed of strings')
        for string in order:
            if not isinstance(string, str):
                raise TypeError('Order should only be composed of strings')
        if self.name == order:
            return self
        for i in range(len(self.der)):
            final_derivative_i = self.der[i].copy()
            for j, variable in enumerate(order):
                index = self.name.index(variable)
                derivative = self.der[i][index].copy()
                final_derivative_i[j] = derivative
            self.der[i] = final_derivative_i
        self.name = order
        return self

    def __add__(self, other):
        """
        Perform addition on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Values to be added to self, which is an AD object.

        Returns
        -------
        AD object representing the result of self+other
        
        Example
        -------
        >>> x = AD(1,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> z = AD(3,1,'z')
        >>> x+y+z
        Numerical Value is:
        [[6.]], 
        Jacobian is:
        [[1. 1. 1.]], 
        Name is:
        ['x', 'y', 'z']
        
        >>> x = AD(1,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> z = AD(3,1,'z')
        >>> v = AD([x+y+z, x+2])
        >>> v
        Numerical Value is:
        [6. 3.], 
        Jacobian is:
        [[1. 1. 1.]
         [1. 0. 0.]], 
        Name is:
        ['x', 'z', 'y']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> f1 = AD([10*x, 10*y])
        >>> f2 = AD([x, y])
        >>> f1+f2+1000
        Numerical Value is:
        [[1022.]
         [1033.]], 
        Jacobian is:
        [[11.  0.]
         [ 0. 11.]], 
        Name is:
        ['x', 'y']
        """
        if isinstance(other, AD):
            names_1 = self.name.copy()
            names_2 = other.name
            value = self.val + other.val
            derivative = self.der.copy()
            for i, name_2 in enumerate(names_2):
                if name_2 in names_1:
                    index_1 = names_1.index(name_2)
                    #     if not isinstance(derivative, np.ndarray):
                      #  derivative = np.array(derivative)
                    #  if not isinstance(other.der, np.ndarray):
                      #   other.der = np.array(other.der)
                    derivative[:, index_1] = derivative[:, index_1] + other.der[:, i]
                else:
                    derivative = np.concatenate((derivative, other.der[:, i].reshape(-1, 1)), axis=1)
                    names_1.append(name_2)
            name = names_1
        else:
            if isinstance(other, np.ndarray):
                if other.shape != self.val.shape:
                    raise ValueError('Cannot add arrays of different size')
                else:
                    for v in other.flatten():
                        if type(v) not in [int, float, np.int, np.int8, np.int16, np.int32, np.int64, \
                                           np.uint8, np.uint16, np.uint32, np.uint64, np.float32, np.float64]:
                            raise TypeError('Array entries must be int or float')
                    else:
                        value = self.val + other
            elif isinstance(other, int) or isinstance(other, float):
                value = self.val + other
            else:
                raise ValueError('Other must be an array, AD, int or float')

            derivative = self.der
            name = self.name
        return AD(value, derivative, name)

    def __radd__(self, other):
        """
        Perform reverse addition on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Values to be added to self, which is an AD object.

        Returns
        -------
        AD object representing the result of self+other
        Example
        -------
        >>> x = AD(1,1,'x') 
        >>> 2+x
        Numerical Value is:
        [[3.]], 
        Jacobian is:
        [[1.]], 
        Name is:
        ['x']        
        
        >>> x = AD(1,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> z = AD([x+y, x])
        >>> 40 + z
        Numerical Value is:
        [[43.]
         [41.]], 
        Jacobian is:
        [[1. 1.]
         [1. 0.]], 
        Name is:
        ['x', 'y']
        """
        new_var = AD(self.val, self.der, self.name)  # create a new variable
        return new_var.__add__(other)

    def __neg__(self):
        """
        Perform negation on AD objects.

        Returns
        -------
        AD object representing the result of -self

        Example
        -------
        >>> x = AD(2,1,'x') 
        >>> -x
        Numerical Value is:
        [[-2.]], 
        Jacobian is:
        [[-1.]], 
        Name is:
        ['x']
        
        >> x = AD(2,1,'x') 
        >> y = AD(3,1,'y') 
        >> z = AD([10*x, 100*y])
        >> -z
        Numerical Value is:
        [[ -20.]
         [-300.]], 
        Jacobian is:
        [[ -10.   -0.]
         [  -0. -100.]], 
        Name is:
        ['x', 'y']
        """
        val = -self.val.copy()
        der = -self.der.copy()
        name = self.name
        return AD(val, der, name)

        # overloading the '-' operator

    def __sub__(self, other):
        """
        Perform subtraction on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Values to be subtracted from self, which is an AD object.

        Returns
        -------
        AD object representing the result of self-other

        Example
        -------
        >>> x = AD(1,1,'x')
        >>> x-3
        Numerical Value is:
        [[-2.]], 
        Jacobian is:
        [[1.]], 
        Name is:
        ['x']
        
        >>> x = AD(1,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> x-y
        Numerical Value is:
        [[-2.]], 
        Jacobian is:
        [[ 1. -1.]], 
        Name is:
        ['x', 'y']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = AD([10*x, 100*y])
        >>> z-2
        Numerical Value is:
        [[ 18.]
         [298.]], 
        Jacobian is:
        [[ 10.   0.]
         [  0. 100.]], 
        Name is:
        ['x', 'y']
        """
        return self.__add__(-other)

    def __rsub__(self, other):
        """
        Perform reverse subtraction on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Values from which self will be subtracted.

        Returns
        -------
        AD object representing the result of other-self
        
        Example
        -------
        >>> x = AD(1,1,'x') 
        >>> 4-x
        Numerical Value is:
        [[3.]], 
        Jacobian is:
        [[-1.]], 
        Name is:
        ['x']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = AD([10*x, 100*y])
        >>> 2-z
       Numerical Value is:
        [[ -18.]
         [-298.]], 
        Jacobian is:
        [[ -10.   -0.]
         [  -0. -100.]], 
        Name is:
        ['x', 'y']
        """
        return -(self.__sub__(other))

    # overloading the '*' operator
    def __mul__(self, other):
        """
        Perform multiplication on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Values multiplying self, which is an AD object.

        Returns
        -------
        AD object representing the result of self*other
        
        Examples
        --------
        >>> x = AD(1,1,'x') 
        >>> x*2
        Numerical Value is:
        [[2.]], 
        Jacobian is:
        [[2.]], 
        Name is:
        ['x']
        
        >>> x = AD(1,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> x*y
        Numerical Value is:
        [[2.]], 
        Jacobian is:
        [[2. 1.]], 
        Name is:
        ['x', 'y']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> f1 = AD([10*x, 10*y])
        >>> f2 = AD([x, y])
        >>> f1*f2
        Numerical Value is:
        [[40.]
         [90.]], 
        Jacobian is:
        [[40.  0.]
         [ 0. 60.]], 
        Name is:
        ['x', 'y']
        """
        if isinstance(other, AD):
            names_1 = self.name.copy()
            names_2 = other.name.copy()
            new_value = self.val.copy() * other.val.copy()
            new_names = names_1 + [name for name in other.name if name not in names_1]
            derivative = self.der.copy()
            for name in new_names:
                # print('Current name', name)
                # print('List of names for 1',names_1)
                # print('List of names for 2', names_2)
                if name in names_1 and name in names_2:
                    index_1 = names_1.index(name)
                    index_2 = names_2.index(name)
                    new_der = self.val.copy() * other.der[:, index_2] + self.der[:, index_1] * other.val
                    derivative[:, index_1] = new_der
                if name in names_1 and name not in names_2:
                    # print('I am here')
                    index_1 = names_1.index(name)
                    new_der = self.der[:, index_1] * other.val
                    # print(new_der)
                    derivative[:, index_1] = new_der
                    # print(derivative)
                if name in names_2 and name not in names_1:
                    # print('Now I am here')
                    index_2 = names_2.index(name)
                    new_der = self.val * other.der[:, index_2]
                    # print(new_der)
                    # print('Before the update', derivative)
                    derivative = np.concatenate((derivative, new_der), axis=1)
            name = new_names
        else:  # one of the coefficients of other is None, it is a constant
            if isinstance(other, np.ndarray):
                for value in other.flatten():
                    if type(value) not in [int, float, np.int, np.int8, np.int16, np.int32, np.int64, \
                                           np.uint8, np.uint16, np.uint32, np.uint64, np.float32, np.float64]:
                            raise TypeError('Arrays must contain only integers or float')
                if self.val.shape[1] == other.shape[0] or self.val.shape == other.shape:
                    new_value = np.dot(self.val, other)
                    derivative = np.dot(self.der, other)
                    name = self.name
                else:
                    raise ValueError('Input dimension mismatch')
            elif isinstance(other, int) or isinstance(other, float):
                new_value = self.val * other
                derivative = self.der * other
                name = self.name
            else:
                raise TypeError('Invalid input type ')
        return AD(new_value, derivative, name)

    def __rmul__(self, other):
        """
        Perform reverse multiplication on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Values to be multiplied by self, which is an AD object.

        Returns
        -------
        AD object representing the result of other*self
        
        
        Examples
        --------
        >>> x = AD(2,1,'x') 
        >>> 2*x
        Numerical Value is:
        [[4.]], 
        Jacobian is:
        [[2.]], 
        Name is:
        ['x']
        
        >>> x = AD(1,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = AD([x+y, x])
        >>> 2*z
        Numerical Value is:
        [[8.]
         [2.]], 
        Jacobian is:
        [[2. 2.]
         [2. 0.]], 
        Name is:
        ['x', 'y']
        """
        return AD(self.val, self.der, self.name).__mul__(other)

    def __truediv__(self, other):  # todo: check for forbidden values
        """
        Perform true division on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Values dividing self, which is an AD object.

        Returns
        -------
        AD object representing the result of self/other
        
        Examples
        --------
        >>> x = AD(4,1,'x') 
        >>> x/2
        Numerical Value is:
        [[2.]], 
        Jacobian is:
        [[0.5]], 
        Name is:
        ['x']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(4,1,'y')
        >>> y/x
        Numerical Value is:
        [[2.]], 
        Jacobian is:
        [[ 1. -2.]], 
        Name is:
        ['y', 'x']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> f1 = AD([10*x, 10*y])
        >>> f2 = AD([2, 1])
        >>> f1/f2
        Numerical Value is:
        [[10.]
         [30.]], 
        Jacobian is:
        [[ 5.  0.]
         [ 0. 10.]], 
        Name is:
        ['x', 'y']
        """
        if isinstance(other, AD):
            names_1 = self.name
            names_2 = other.name
            for val in other.val:
                if val == 0:
                    raise ZeroDivisionError
            new_value = self.val / other.val  # numpy array multiplication
            new_names = names_1 + [name for name in other.name if
                                   name not in names_1]  # we should raise an error when dividing by 0
            derivative = self.der.copy()
            for name in new_names:
                if name in names_1 and name in names_2:
                    index_1 = names_1.index(name)
                    index_2 = names_2.index(name)
                    new_der = (self.der[:, index_1] * other.val - other.der[:,
                                                                  index_2] * self.val) / other.val ** 2  # with scalars
                    derivative[:, index_1] = new_der
                if name in names_1 and name not in names_2:
                    # print('I am here')
                    index_1 = names_1.index(name)
                    new_der = self.der[:, index_1] / other.val
                    derivative[:, index_1] = new_der
                if name in names_2 and name not in names_1:
                    index_2 = names_2.index(name)
                    new_der = -self.val * other.der[:, index_2] / other.val ** 2
                    derivative = np.concatenate((derivative, new_der), axis=1)
            name = new_names
        else:
            if isinstance(other, np.ndarray):
                if other.shape != self.val.shape:
                    raise TypeError('Input dimension mismatch')
                for v in other.flatten():
                    if type(v) not in [int, float, np.int, np.int8, np.int16, np.int32, np.int64, \
                                       np.uint8, np.uint16, np.uint32, np.uint64, np.float32, np.float64]:
                        raise TypeError('Array entries must be int or float')
                if np.min(np.abs(other)) == 0:
                    raise ZeroDivisionError
                new_value = self.val / other
                derivative = self.der / other
                name = self.name
            elif isinstance(other, int) or isinstance(other, float):
                if other == 0:
                    raise ZeroDivisionError
                new_value = self.val / other
                derivative = self.der / other
                name = self.name
            else:
                raise TypeError('Wrong input type')
        return AD(new_value, derivative, name)

    def __rtruediv__(self, other):
        """
        Perform reverse true division on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Values to be divided by self, which is an AD object.

        Returns
        -------
        AD object representing the result of other/self
        
        Examples
        --------
        >>> x = AD(10,1,'x') 
        >>> 2/x
        Numerical Value is:
        [[0.2]], 
        Jacobian is:
        [[-0.02]], 
        Name is:
        ['x']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = x+y
        >>> 1/z
        Numerical Value is:
        [[0.2]], 
        Jacobian is:
        [[-0.04 -0.04]], 
        Name is:
        ['x', 'y']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = AD([x, y])
        >>> 1/z
        Numerical Value is:
        [[0.5       ]
         [0.33333333]], 
        Jacobian is:
        [[-0.25       -0.        ]
         [-0.         -0.11111111]], 
        Name is:
        ['x', 'y']
        """
        if np.abs(np.min(self.val)) == 0:
            raise ZeroDivisionError
        if not isinstance(other, int) and not isinstance(other, float):
            raise TypeError('Invalid input type')
        new_val = other / self.val
        new_der = -other * self.der / self.val ** 2
        names = self.name
        return AD(new_val, new_der, names)


    def __pow__(self, n):
        """
        Raise an AD object to the power of n.

        Parameters
        ----------
        n : int, float or AD
            Exponent to which self will be raised. Self is an AD object.

        Returns
        -------
        AD object representing the result of self**other
        
        Examples
        --------
        >>> x = AD(2,1,'x') 
        >>> x**3
        Numerical Value is:
        [[8.]], 
        Jacobian is:
        [[12.]], 
        Name is:
        ['x']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y')
        >>> x**y
        Numerical Value is:
        [[8.]], 
        Jacobian is:
        [[12.          5.54517744]], 
        Name is:
        ['x', 'y']

        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = x+y
        >>> z**2
        Numerical Value is:
        [[25.]], 
        Jacobian is:
        [[10. 10.]], 
        Name is:
        ['x', 'y']

        """
        if isinstance(n, float) or isinstance(n, int):
            n = float(n)  # n is an int/float
            value = self.val.copy()
            derivative = self.der.copy()
            names = self.name
            if (value <= 0).any() and 0 < n < 1:
                raise ValueError('Illegal value and exponent')
            if (value == 0).any() and n < 0:
                raise ZeroDivisionError
            new_val = value ** n
            for i, name in enumerate(names):
                derivative[:, i] = n * self.der[:, i] * value ** (n - 1)
            return AD(new_val, derivative, names)

        elif isinstance(n, AD):  # n is an AD object
            value_base = self.val
            value_exponent = n.val
            if not self.val.shape == n.val.shape:
                for v in n.val.flatten():
                    if type(v) not in [int, float, np.int, np.int8, np.int16, np.int32, np.int64, \
                                       np.uint8, np.uint16, np.uint32, np.uint64, np.float32, np.float64]:
                        raise TypeError('Array entries must be int or float')
            for val_base, val_exponent in zip(value_base, value_exponent):
                if val_base < 0 and 0 < val_exponent < 1:
                    raise ValueError('Illegal value and exponent')
                if val_base == 0 and val_exponent < 0:
                    raise ZeroDivisionError
            new_val = value_base ** value_exponent
            names_1 = self.name
            names_2 = n.name
            new_names = names_1 + [name for name in names_2 if name not in names_1]
            derivative = self.der.copy()
            for i, name in enumerate(new_names):
                if name in names_1 and name in names_2:
                    index_2 = names_2.index(name)
                    new_der = (n.der[:, index_2] * np.log(value_base) + value_exponent * derivative[:,
                                                                                           i] / value_base) * new_val
                    derivative[:, i] = new_der
                if name in names_1 and name not in names_2:
                    index_1 = names_1.index(name)
                    new_der = derivative[:, index_1] * value_exponent * value_base ** (value_exponent - 1)
                    derivative[:, index_1] = new_der
                if name in names_2 and name not in names_1:
                    index_2 = names_2.index(name)
                    new_der = n.der[:, index_2] * np.log(value_base) * new_val
                    derivative = np.concatenate((derivative, new_der), axis=1)
            return AD(new_val, derivative, new_names)
        else:
            raise TypeError('Invalid Input Type for the exponent')


    def __rpow__(self, other):
        """
        Raise a number to the power of an AD object.

        Parameters
        ----------
        other : int or float
            Exponent to which self will be raised. Self is an AD object.

        Returns
        -------
        AD object representing the result of other**self
        
        Examples
        --------
        >>> x = AD(2,1,'x') 
        >>> 3**x
        Numerical Value is:
        [[9.]], 
        Jacobian is:
        [[6.23832463]], 
        Name is:
        ['x']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = x+y
        >>> 2**z
        Numerical Value is:
        [[32.]], 
        Jacobian is:
        [[51.5020132 51.5020132]], 
        Name is:
        ['x', 'y']
        """
        value = self.val
        if isinstance(other, int) or isinstance(other, float):
            if other < 0:
                raise ValueError('Inconsistent value found for the base')
            if other == 0 and np.min(value) < 0:
                raise ZeroDivisionError
            der = self.der
            name = self.name
            new_value = other ** value
            new_der = der * np.log(other) * new_value
        elif isinstance(other, np.ndarray):
            if other.shape != value.shape:
                raise ValueError('Invalid dimension')
            for val in other:
                if not isinstance(val, int) and not isinstance(val, float):
                    raise TypeError('Invalid input type')
                if val < 0:
                    raise ValueError('Inconsistent value found for the base')
                if val == 0 and np.min(value) == 0:
                    raise ZeroDivisionError
            der = self.der
            name = self.name
            new_value = other ** value
            new_der = der * np.log(value) * new_value
        else:
            raise TypeError('Invalid input type')
        return AD(new_value, new_der, name)


    def __lt__(self, other):
        """
        Perform "less than" comparison on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Value to compare with self, which is an AD object.

        Returns
        -------
        Boolean representing the result of self < other
        
        Examples
        --------
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = x+y
        >>> v = 2*(x+y)
        >>> u < v
        True
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = AD([x+y, x+y])
        >>> v = AD([x+y, 2*(x+y)])
        >>> u < v
        True
        """
        if isinstance(other, AD):
            if self.val.shape != other.val.shape:
                raise AttributeError('Incoherent dimension input')
            for val, other_value in zip(self.val, other.val):
                if val >= other_value:
                    return False
            return True
        elif isinstance(other, int) or isinstance(other, float):
            for value in self.val:
                if value >= other:
                    return False
            return True
        else:
            raise TypeError('Invalid input type')



    def __gt__(self, other):
        """
        Perform "greater than" comparison on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Value to compare with self, which is an AD object.

        Returns
        -------
        Boolean representing the result of self > other
        
        
        Examples
        --------
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = x+y
        >>> v = 2*(x+y)
        >>> v > u
        True
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = AD([x+y, x+y])
        >>> v = AD([x+y, 2*(x+y)])
        >>> v > u
        True
        """
        if isinstance(other, AD):
            return other.__lt__(self)
        elif isinstance(other, int) or isinstance(other, float):
            for value in self.val:
                if value <= other:
                    return False
            return True
        else:
            raise TypeError('Invalid input type')


    def __eq__(self, other):
        """
        Perform "equality" comparison on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Value to compare with self, which is an AD object.

        Returns
        -------

        AD object representing the result of self == other

        Boolean representing the result of self <= other
        
        Examples
        --------
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = x+y
        >>> v = 2*(x+y)
        >>> u <= v , u <= v/2 
        (True, True)
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = AD([x+y, x+y])
        >>> v = AD([x+y, 2*(x+y)])
        >>> u <= v
        True

        """
        return not bool(self.__lt__(other)) and not bool(self.__gt__(other))


    def __le__(self, other):
        """
        Perform "less or equal than" comparison on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Value to compare with self, which is an AD object.

        Returns
        -------

        AD object representing the result of self <= other

        Boolean representing the result of self >= other
        
        Examples
        --------
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = x+y
        >>> v = 2*(x+y)
        >>> v >= u, v/2 >= u
        (True, True)
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = AD([x+y, x+y])
        >>> v = AD([x+y, 2*(x+y)])
        >>> v >= u
        True

        """
        print('Eq', bool(self.__eq__(other)))
        print('Le', bool(self.__lt__(other)))
        return (bool(self.__lt__(other)) or bool(self.__eq__(other)))

    def __ge__(self, other):
        """
        Perform "greater or equal than" comparison on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Value to compare with self, which is an AD object.

        Returns
        -------

        AD object representing the result of self >= other

        Boolean representing the result of self == other
        
        Examples
        --------
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = x+y
        >>> v = 2*(x+y)
        >>> v == u, v/2 == u
        (False, True)
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = AD([x+y, x+y])
        >>> v = AD([x+y, x+y])
        >>> v == u
        True

        """
        # raises an error when the two objects do not have the same attributes
        return self.__gt__(other) or self.__eq__(other)



    def __ne__(self, other):
        """
        Perform "inequality" comparison on an AD object.

        Parameters
        ----------
        other : int, float, np.ndarray, list or AD
            Value to compare with self, which is an AD object.

        Returns
        -------
        AD object representing the result of self != other
        
        Examples
        --------
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = x+y
        >>> v = 2*(x+y)
        >>> v != u, v/2 != u
        (True, False)
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(2,1,'y') 
        >>> u = AD([x+y, x+y])
        >>> v = AD([x+y, x+y])
        >>> v != u
        False
        """
        return not self.__eq__(other)


    def tan(self):
        """
        Compute the tangent of an AD object.

        Returns
        -------
        AD object representing the result of tan(self)
        
        Examples
        --------
        >>> x = AD(np.pi/4,1,'x') 
        >>> x.tan()
        Numerical Value is:
        [[1.]], 
        Jacobian is:
        [[2.]], 
        Name is:
        ['x']
        
        >>> x = AD(np.pi/4,1,'x') 
        >>> y = AD(np.pi/4,1,'y')
        >>> z = AD([x,y])
        >>> z.sin()
        Numerical Value is:
        [[1.]
         [1.]], 
        Jacobian is:
        [[2. 0.]
         [0. 2.]], 
        Name is:
        ['x', 'y']
        """
        nonpoints = map(lambda x: ((x / np.pi) - 0.5) % 1 == 0.00, self.val)
        if any(nonpoints):
            raise ValueError("Math error, Tangent cannot handle i*0.5pi ")
        val = np.tan(self.val)
        der = np.multiply(1 / np.power(np.cos(self.val), 2), self.der)
        return AD(val, der, self.name)


    def sin(self):
        """
        Compute the sine of an AD object.

        Returns
        -------
        AD object representing the result of sin(self)
        
        Examples
        --------
        >>> x = AD(np.pi/4,1,'x') 
        >>> x.sin()
        Numerical Value is:
        [[0.70710678]], 
        Jacobian is:
        [[0.70710678]], 
        Name is:
        ['x']
        
        >>> x = AD(np.pi/4,1,'x') 
        >>> y = AD(np.pi/4,1,'y')
        >>> z = AD([x,y])
        >>> z.sin()
        Numerical Value is:
        [[0.70710678]
         [0.70710678]], 
        Jacobian is:
        [[0.70710678 0.        ]
         [0.         0.70710678]], 
        Name is:
        ['x', 'y']
        """
        val = np.sin(self.val)
        der = np.cos(self.val) * self.der
        return AD(val, der, self.name)

    def cos(self):
        """
        Compute the cosine of an AD object.

        Returns
        -------
        AD object representing the result of cos(self)
        
        Examples
        --------
        >>> x = AD(np.pi/4,1,'x') 
        >>> x.cos()
        Numerical Value is:
        [[0.70710678]], 
        Jacobian is:
        [[-0.70710678]], 
        Name is:
        ['x']
        
        >>> x = AD(np.pi/4,1,'x') 
        >>> y = AD(np.pi/4,1,'y')
        >>> z = AD([x,y])
        >>> z.cos()
        Numerical Value is:
        [[0.70710678]
         [0.70710678]], 
        Jacobian is:
        [[-0.70710678 -0.        ]
         [-0.         -0.70710678]], 
        Name is:
        ['x', 'y']
        """
        val = np.cos(self.val)
        der = -np.sin(self.val) * self.der
        return AD(val, der, self.name)

    def exp(self):
        """
        Compute the exponential of an AD object.

        Returns
        -------
        AD object representing the result of exp(self)
        
        Examples
        --------
        >>> x = x = AD(2,1,'x') 
        >>> x.exp() 
        Numerical Value is:
        [[7.3890561]], 
        Jacobian is:
        [[7.3890561]], 
        Name is:
        ['x']

        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = AD([x,y])
        >>> z.exp()
        Numerical Value is:
        [[ 7.3890561 ]
         [20.08553692]], 
        Jacobian is:
        [[ 7.3890561   0.        ]
         [ 0.         20.08553692]], 
        Name is:
        ['x', 'y']
        """
        #val = np.exp(self.val)
        #der = np.multiply(np.exp(self.val), self.der)
        #return AD(val, der, self.name) # return np.exp(1)**self
        return self.__rpow__(np.exp(1))


    def ln(self):
        """
        Compute the natural logarithm of an AD object.

        Returns
        -------
        AD object representing the result of ln(self)
        
        Examples
        --------
        >>> x = x = AD(2,1,'x') 
        >>> x.ln() 
        Numerical Value is:
        [[0.69314718]], 
        Jacobian is:
        [[0.5]], 
        Name is:
        ['x']
        
        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = AD([x,y])
        >>> z.ln()
        Numerical Value is:
        [[0.69314718]
         [1.09861229]], 
        Jacobian is:
        [[0.5        0.        ]
         [0.         0.33333333]], 
        Name is:
        ['x', 'y']
        """

        for val in self.val:
            if val <= 0:
                raise ValueError("Cannot take natural log of zero or negative values")
        val = np.log(self.val)
        der = self.der / self.val
        return AD(val, der, self.name)


    def ln_base(self, base):
        """
        Compute the base-specific logarithm of an AD object.

        Returns
        -------
        AD object representing the result of ln_base(self)
        
        Examples
        --------
        >>> x = x = AD(8,1,'x') 
        >>> x.ln_base(2) 
        Numerical Value is:
        [[3.]], 
        Jacobian is:
        [[0.18033688]], 
        Name is:
        ['x']
        
        >>> x = AD(8,1,'x') 
        >>> y = AD(32,1,'y') 
        >>> z = AD([x*y,y])
        >>> z.ln_base(2)
        Numerical Value is:
        [[8.]
         [5.]], 
        Jacobian is:
        [[0.18033688 0.36067376]
         [0.         0.04508422]], 
        Name is:
        ['x', 'y']
        """
        return self.ln()/np.log(base)


    def sinh(self):  # hyperbolic sin
        """
        Compute the hyperbolic sine of an AD object.

        Returns
        -------
        AD object representing the result of sinh(self)
        
        Examples
        --------
        >>> x = x = AD(2,1,'x') 
        >>> x.sinh() 
        Numerical Value is:
        [[3.62686041]], 
        Jacobian is:
        [[3.76219569]], 
        Name is:
        ['x']

        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = AD([x,y])
        >>> z.sinh()
        Numerical Value is:
        [[ 3.62686041]
         [10.01787493]], 
        Jacobian is:
        [[ 3.76219569  0.        ]
         [ 0.         10.067662  ]], 
        Name is:
        ['x', 'y']
        """
        # d/dx (sinh x) = cosh x
        # sinh x = (e^x - e^(-x))/2  range (-inf, inf)
        # val = np.multiply(.5, (np.exp(self.val) - np.exp(np.multiply(-1, self.val))))
        val = np.sinh(self.val)
        der = np.cosh(self.val) * self.der
        return AD(val, der, self.name)

    def cosh(self):  # hyperbolic cos
        """
        Compute the hyperbolic cosine of an AD object.

        Returns
        -------
        AD object representing the result of cosh(self)
        
        Examples
        --------
        >>> x = x = AD(2,1,'x') 
        >>> x.cosh() 
        Numerical Value is:
        [[3.76219569]], 
        Jacobian is:
        [[3.62686041]], 
        Name is:
        ['x']

        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = AD([x,y])
        >>> z.cosh()
        Numerical Value is:
        [[ 3.62686041]
         [10.01787493]], 
        Jacobian is:
        [[ 3.76219569  0.        ]
         [ 0.         10.067662  ]], 
        Name is:
        ['x', 'y']
        """
        # d/dx (cosh x) = sinh x
        # cosh x = (e^x + e^(-x))/2  range (-inf, inf)
        # val = np.multiply(.5, (np.exp(self.val) + np.exp(np.multiply(-1, self.val))))
        val = np.cosh(self.val)
        der = np.sinh(self.val) * self.der
        return AD(val, der, self.name)


    def tanh(self):  # hyperbolic tan
        """
        Compute the hyperbolic tangent of an AD object.

        Returns
        -------
        AD object representing the result of tanh(self)
        
        Examples
        --------
        >>> x = x = AD(2,1,'x') 
        >>> x.tanh() 
        Numerical Value is:
        [[0.96402758]], 
        Jacobian is:
        [[0.07065082]], 
        Name is:
        ['x']

        >>> x = AD(2,1,'x') 
        >>> y = AD(3,1,'y') 
        >>> z = AD([x,y])
        >>> z.tanh()
       Numerical Value is:
        [[0.96402758]
         [0.99505475]], 
        Jacobian is:
        [[0.07065082 0.        ]
         [0.         0.00986604]], 
        Name is:
        ['x', 'y']
        """
        # d/dx (tanh x) = (sech x)^2 = 1/((cosh x)^2)
        # tanh x = (e^x - e^(-x)) / (e^x + e^(-x))       range (-inf, inf)
        val = np.tanh(self.val)
        der = (1 / np.power(np.cosh(self.val), 2)) * self.der
        # return self.sinh()/self.cosh()
        return AD(val, der, self.name)


    def arcsin(self):
        """
        Compute the arcsine (inverse of sine) of an AD object.

        Returns
        -------
        AD object representing the result of arcsin(self)
        
        Examples
        --------
        >>> x = AD(-0.5,1,'x') 
        >>> x.arcsin()
        Numerical Value is:
        [[-0.52359878]], 
        Jacobian is:
        [[1.15470054]], 
        Name is:
        ['x']
        
        >>> x = AD(-0.5,1,'x') 
        >>> y = AD(0.5,1,'y')
        >>> z = AD([x,y])
        >>> z.arcsin()
        Numerical Value is:
        [[-0.52359878]
         [ 0.52359878]], 
        Jacobian is:
        [[1.15470054 0.        ]
         [0.         1.15470054]], 
        Name is:
        ['x', 'y']
        """
        if ((self.val <= -1).any() or (self.val>=1).any()):
           raise ValueError("Cannot take derivative of arcsin of value outside of range (-1, 1)")
        val = np.arcsin(self.val)
        # der = (1/(np.sqrt(1 - np.power(self.val, 2)))) * self.der
        der = self.der * ((1 - self.val ** 2) ** (-0.5))
        return AD(val, der, self.name)


    def arccos(self):
        """
        Compute the arccosine (inverse of cosine) of an AD object.

        Returns
        -------
        AD object representing the result of arccos(self)
        
        Examples
        --------
        >>> x = AD(-0.5,1,'x') 
        >>> x.arccos()
        Numerical Value is:
        [[2.0943951]], 
        Jacobian is:
        [[-1.15470054]], 
        Name is:
        ['x']
        
        >>> x = AD(-0.5,1,'x') 
        >>> y = AD(0.5,1,'y')
        >>> z = AD([x,y])
        >>> z.arccos()
        Numerical Value is:
        [[2.0943951 ]
         [1.04719755]], 
        Jacobian is:
        [[-1.15470054 -0.        ]
         [-0.         -1.15470054]], 
        Name is:
        ['x', 'y']
        """
        if (self.val <= -1).any() or (self.val >= 1).any():
            raise ValueError("Cannot take derivative of arcsin of value outside of range (-1, 1)")
        val = np.arccos(self.val)
        # der = -(1/(np.sqrt(1 - np.power(self.val, 2)))) * self.der
        der = -self.der * ((1 - self.val ** 2) ** (-0.5))
        return AD(val, der, self.name)


    def arctan(self):
        """
        Compute the arctangent (inverse of tangent) of an AD object.

        Returns
        -------
        AD object representing the result of arctan(self)
        
        Examples
        --------
        >>> x = AD(-0.5,1,'x') 
        >>> x.arctan()
        Numerical Value is:
        [[-0.46364761]], 
        Jacobian is:
        [[0.8]], 
        Name is:
        ['x']
        
        >>> x = AD(-0.5,1,'x') 
        >>> y = AD(0.5,1,'y')
        >>> z = AD([x,y])
        >>> z.arctan()
        Numerical Value is:
        [[-0.46364761]
         [ 0.46364761]], 
        Jacobian is:
        [[0.8 0. ]
         [0.  0.8]], 
        Name is:
        ['x', 'y']
        """
        val = np.arctan(self.val)
        der = self.der * (1 + self.val ** 2) ** (-1)
        return AD(val, der, self.name)


    def logistic(self):
        """
        Apply the sigmoid function to an AD object.

        The sigmoid function of x is defined as:

            sigmoid(x) =  1/(1+e**(-x))

        Returns
        -------
        AD object representing the result of sigmoid(self)
        
        Examples
        --------
        >>> x = AD(-0.5,1,'x') 
        >>> x.logistic()
        Numerical Value is:
        [[0.37754067]], 
        Jacobian is:
        [[0.23500371]], 
        Name is:
        ['x']
        
        >>> x = AD(-0.5,1,'x') 
        >>> y = AD(0.5,1,'y')
        >>> z = AD([x,y])
        >>> z.logistic()
        Numerical Value is:
        [[0.37754067]
         [0.62245933]], 
        Jacobian is:
        [[0.23500371 0.        ]
         [0.         0.23500371]], 
        Name is:
        ['x', 'y']
        """
        # assuming logistic function = sigmoid function = 1/(1+e^(-x))
        val = 1 / (1 + np.exp(-self.val))
        der = self.der * np.exp(-self.val) / ((1 + np.exp(-self.val)) ** 2)
        return AD(val, der, self.name)


    def sqrt(self):
        """
        Compute the square root of an AD object.

        Returns
        -------
        AD object representing the result of sigmoid(self)
        
        Examples
        --------
        >>> x = AD(25,1,'x') 
        >>> x.sqrt()
        Numerical Value is:
        [[5.]], 
        Jacobian is:
        [[0.1]], 
        Name is:
        ['x']
        
        >>> x = AD(4,1,'x') 
        >>> y = AD(9,1,'y')
        >>> z = AD([x,y])
        >>> z.sqrt()
        Numerical Value is:
        [[2.]
         [3.]], 
        Jacobian is:
        [[0.25       0.        ]
         [0.         0.16666667]], 
        Name is:
        ['x', 'y']
        """

        # if (self.val < 0).any():
          #  raise ValueError('Square root should only be considered for positive numbers')
        # new_val = np.sqrt(self.val.copy())
        # if (self.val == 0).any():
          #  raise ValueError('The derivative of the square root can only be computed for strictly positive numbers')
        # new_der = 1 / 2 * self.der * (self.val ** -0.5)
        #  return AD(new_val, new_der, self.name)
        return self.__pow__(0.5)



import numpy as np
from AD.ADmulti import AD
import inspect
import math

names = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
         'w']
# todo: names inside the class and initialize with numbers a1


class Optimizer:

    """This class implements different types of Optimizers. Their common paradigm is gradient descent. The
    way these optimizers differ is the way the updates are performed, and whether they include momentum
    control or second moment of the gradient control."""

    def __init__(self, f, max_iter=1e6, init_points=None, dimensions=0, batch_size=0, tolerance=1e-6, random_restarts=2, beta_1=0.9, beta_2=0.999,
                 step_size=1e-3, epsilon=1e-8, tuning=False, quadratic_matrix=None, max_epochs=10, verbose=0):
        """Initializes an object of the class Optimizer, allowing to optimize a function based on a descent method you chose.
        INPUTS
        =======
        f: function, takes arbitrary number of scalar inputs, or a vector input.
            The function we wish to minimize
        init_points: list, optional, default value is None
            The initial points from which you want to perform your descent. If it not specified by the user,
            the algorithm will create some. By default, these points are drawn from a multivariate gaussian centered
            at the origin of the input space.
        dimensions: int, default value is 0
            This allow the optimizer to understand that the input is a vector, and the user specifies the size of the vector
            he wishes to get back from the optimizer. Is not necessary when init_points is input.
        batch_size: int, default value is 0
            This parameter allow to perform stochastic gradient descent, ie gradient computation based on a subset of the data
            and not based on the entire data. This value should be ess that the size of your input array
        max_iter: int, optional, default value is 1e6
            The max number of updates you wish to perform during your descent algorithm
        tolerance: int, optional, default value is 1e-6
            Threshold on updates performed on the weight parameter. This allows to perform early stopping for the
            algorithm in zones where the weights are not updated anymore
        random_restarts: int, optional, default value is 2
            Number of starts the user wants to restart the descent from a different random point. The final optimal
            point selected will be the best of the different (random_restarts) points
        beta_1: float, optional, default value is 0.9
            Momentum on the first moment of the gradient. Default value selected accordingly to Tensorflow's algorithms
        beta_2: float, optional, default value is 0.99
            Momentum on the second moment of the gradient. Default value selected accordingly to Tensorflow's algorithms
        step_size: float, optional, default value is 1e-3
            Static step size wanted during the descent.
        epsilon: float, optional, default value is 1e-8
            Noise introduced in order to prevent the involved denominators from being 0
        tuning: boolean, optional, default value is False
            Future feature to be developed. This will allow to tune the initial point before starting the descent.
            The algorithm leveraged is simulated annealing, which allow to sample points from the 'interesting' zones
            in the optimization landscape of a function. Right now, it is efficient for quadratic functions (that have
            the form x^T*A*x
        quadratic_matrix: numpy matrix, optional, default value is None
            The quadratic matrix for the definition of the quadratic form. This will allow to efficiently sample
            points that are interesting in the optimization landscape of the quadratic form.
        max_epochs: int, optional, default value is 100
            The number of times you wish to perform simulated annealing algorithm in order to get a good starting point.
            The higher the better
        verbose: integer, optional, default value is 0
            Value between 0 and 2 describing the amount of information we wish to be print during the optimization
            algorithm, for debugging issues

        RETURNS
        ========
        An instance of the class Optimizer

        EXAMPLES
        =========
        Without sampling and when specifying initialization points
        >>> f = lambda x : x
        >>> max_iter = 10
        >>> init_points = [1]
        >>> tolerance = 1
        >>> random_restarts = 2
        >>> beta_1 = 0.5
        >>> beta_2 = 0.1
        >>> step_size = 1e-2
        >>> epsilon = 1e-4
        >>> tuning = False
        >>> verbose = 0
        >>> opt = Optimizer(f=f, max_iter=max_iter, init_points=init_points, tolerance=tolerance, random_restarts=random_restarts, beta_1=beta_1, beta_2=beta_2, step_size=step_size, epsilon=epsilon, tuning=tuning, verbose=verbose)
        Without sampling, just specifying the function and without specifying input points
        >>> f = lambda x : x
        >>> opt = Optimizer(f=f)
        With sampling, specifying the quadratic matrix of the quadratic form
        >>> f = lambda x, y, z: x ** 2 + y ** 2 + z ** 2
        >>> opt = Optimizer(f, quadratic_matrix=np.eye(3), tuning=True)
        With sampling, specifying the quadratic matrix of the quadratic form with verbose in order to debug
        >>> f = lambda x, y, z: x ** 2 + y ** 2 + z ** 2
        >>> opt = Optimizer(f, quadratic_matrix=np.eye(3), tuning=True, verbose=2)
        """
        self.init_function = f
        self.trace_values = []
        self.trace_gradients = []
        self.trace = []
        if not isinstance(max_iter, int) and not isinstance(max_iter, float):
            raise TypeError('Max iter should be an integer or a float')
        if max_iter < 0:
            raise ValueError('Max iter should be a > 0 integer')
        self.max_iter = max_iter
        if isinstance(init_points, int) or isinstance(init_points, float):
            init_points = [init_points]
        if init_points is not None and not isinstance(init_points, list):
            raise TypeError('Initial points should be a list')
        if not isinstance(batch_size, int):
            raise ValueError('Batch Size should be an integer')
        if not isinstance(dimensions, int):
            raise ValueError('Dimension should be an integer')
        if init_points is None:
            self.dimension = dimensions
            if self.dimension > 0:
                self.vectorize = True
                if batch_size > 0 and batch_size > dimensions:
                    raise ValueError('Your batch size should be smaller than your input size')
                self.batch_size = batch_size
            else:
                self.vectorize = False
        if init_points is not None:
            try:
                f(*init_points)
                self.vectorize = False
            except TypeError:
                try:
                    f(np.array(init_points))
                    self.vectorize = True
                    self.dimension = len(init_points)
                    if batch_size > 0 and batch_size > self.dimension:
                        raise ValueError('Your batch size should be smaller than your input size')
                    self.batch_size = batch_size
                except TypeError:
                    raise ValueError('Please enter valid input points')
        self.init = init_points
        if not isinstance(tolerance, float) and not isinstance(tolerance, int):
            raise TypeError('Tolerance should be an integer or a float')
        if tolerance < 0:
            raise ValueError('Tolerance should be > 0')
        self.tol = tolerance
        if not isinstance(random_restarts, int):
            raise TypeError('Random restarts should be an integer')
        if random_restarts < 0:
            raise ValueError('Random restarts should be > 0')
        self.restarts = random_restarts
        if not isinstance(step_size, float) and not isinstance(step_size, int):
            raise TypeError('Step size should be an integer or a float')
        if step_size < 0:
            raise ValueError('Step size should be > 0')
        self.lr = step_size
        if not isinstance(beta_1, int) and not isinstance(beta_1, float):
            raise TypeError('Beta_1 should be an integer or a float')
        if beta_1 > 1 or beta_1 < 0:
            raise ValueError('Beta_1 should be an integer or a float between 0 and 1')
        self.beta1 = beta_1
        if not isinstance(beta_2, int) and not isinstance(beta_2, float):
            raise TypeError('Beta_2 should be an integer or a float')
        if beta_2 > 1 or beta_2 < 0:
            raise ValueError('Beta_1 should be an integer or a float between 0 and 1')
        self.beta2 = beta_2
        if not isinstance(epsilon, int) and not isinstance(epsilon, float):
            raise TypeError('Epsilon should be an integer or a float')
        self.eps = epsilon
        if not isinstance(tuning, bool):
            raise TypeError('Tuning should be a boolean')
        self.sampling = tuning
        self.global_optimizer = None
        if not self.sampling and quadratic_matrix is not None:
            raise ValueError('You should set tuning to true when inputing a covariance matrix')
        if quadratic_matrix is not None and not isinstance(quadratic_matrix, np.ndarray):
            raise TypeError('Quadratic matrix should be an array')
        if quadratic_matrix is not None and len(quadratic_matrix.shape) < 2:
            raise ValueError('Quadratic matrix should be a 2D array')
        if quadratic_matrix is not None and init_points is not None:
            if quadratic_matrix.shape[0] != len(init_points):
                raise ValueError('Input dimensions not coherent for cov matrix and init points')
        self.covariance = quadratic_matrix
        if not isinstance(max_epochs, int):
            raise TypeError('Epochs should be an integer')
        if max_epochs < 0:
            raise ValueError('The number of epochs should be an integer > 0')
        self.max_epochs = max_epochs
        if not isinstance(verbose, int):
            raise TypeError('Verbose should be an integer')
        if verbose < 0:
            raise ValueError('The verbose should be an integer > 0')
        self.verbose = verbose
        self.num_iterations = []

    def __str__(self):
        if self.global_optimizer:
            return 'Adam optimizer, which found an optimal weight point of ' + str(
                self.global_optimizer) + ' the value of the objective function at this point is ' + str(
                self.init_function(*self.global_optimizer))
        else:
            return 'Adam optimizer, not yet fitted'

    def produce_random_points(self):
        """This function allows to produce random initialization points in order to start the descent algorithm when
        the user does not wish to specify some initialization points. This function infer the dimensionality of
        the inputs and returns points sampled from a gaussian distribution. Returns, init_point, array of shape (1, n_args)
        INPUTS
        =======
        self: an adam instance of the optimizer

        RETURNS
        ========
        final_values: list of values
           Has the form (x, y, z) where this point is the best point found across the different random restarts

        EXAMPLES
        =========
        >>> f = lambda x: x**2+y**2+(z-2)**2
        >>> opt = Optimizer(f)
        >>> init = opt.produce_random_points()
        >>> [-1.4738622, 1.0198161, -0.2515112]"""
        if self.vectorize:
            if self.dimension > 0:  # the user did not specify any input points
                dimensionality = self.dimension
                sample_produced = np.random.multivariate_normal(np.zeros(dimensionality), np.eye(dimensionality))
        else:
            n_args = len(inspect.getargspec(self.init_function).args)
            test_dim_input = np.ones(n_args).reshape(n_args, 1)
            dimensionality = len(test_dim_input)
            sample_produced = np.random.multivariate_normal(np.zeros(dimensionality), np.eye(dimensionality))
        return sample_produced

    def descent(self):
        """Run the gradient descent algorithm to minimize a function, with the Adam modification for the adaptation
        of the learning rate, or the sgd optimizer, depending on the values of beta1 and beta2.
        INPUTS
        =======
        self: an adam instance of the optimizer

        RETURNS
        ========
        final_values: list of values
           Has the form (x, y, z) where this point is the best point found across the different random restarts

        EXAMPLES
        =========
        >>> f = lambda x: x**2+y**2+(z-2)**2
        >>> adam = Adam(f, random_restarts=10)
        >>> adam.descent()
        >>> print(adam.global_optimizer)
        [-3.1341273654385723e-06, 1.9563233193181366e-59, 1.9998541875049667]"""
        accumulator = []
        final_values = []
        for _ in range(self.restarts):
            if self.init is not None:
                init_point = self.init
            else:
                init_point = self.produce_random_points()
            if self.sampling:
                init_point = self.annealing(init_point)
            w = [AD(init_point[i], 1, name='w'+str(i)) for i in range(len(init_point))] # w should be an AD variable
            if self.vectorize:
                w = np.array(w)
            m = 0
            v = 0  # same as tf
            self.trace_values = []
            self.trace = []
            for t in range(1, int(self.max_iter)):
                if not self.vectorize:
                    AD_function = self.init_function(*w)
                else:
                    AD_function = self.init_function(w)
                self.trace_values.append(AD_function.val[0])
                gradient = AD_function.der
                self.trace_gradients.append(gradient)
                if self.beta1 != 0 and self.beta2 != 0:  # want to perform a adam gradient descent
                    m = self.beta1 * m + (1 - self.beta1) * gradient
                    v = self.beta2 * v + (1 - self.beta2) * np.power(gradient, 2)
                    m_hat = m / (1 - np.power(self.beta1, t))  # unbiased estimator
                    v_hat = v / (1 - np.power(self.beta2, t))  # unbiased estimator
                    update = self.lr * m_hat / (np.sqrt(v_hat) + self.eps)
                if self.beta1 == 0 and self.beta2 != 0:  # want to perform RMS prop
                    v = self.beta2 * v + (1 - self.beta2) * np.power(gradient, 2)
                    v_hat = v / (1 - np.power(self.beta2, t))  # unbiased estimator
                    update = self.lr * gradient / (np.sqrt(v_hat) + self.eps)
                elif self.beta1 == 0 and self.beta2 == 0:  # want to perform single sgd/gd without momentum
                    update = self.lr * gradient
                if np.sqrt(np.sum(update ** 2)) < self.tol:
                    break
                w = w - update  # weight update
                w = w[0]
                self.trace.append(np.array([w.val[0][0] for w in w]))  # fix we have made, to be checked
            self.num_iterations.append(t)
            final_value = [float(w.val) for w in w]
            if not self.vectorize:
                accumulator.append(self.init_function(*final_value))
            else:
                accumulator.append(self.init_function(np.array(final_value)))
            final_values.append(final_value)
            if self.verbose >= 2:
                print('Descent algorithm performed, the optimal point found is ' + str(
                    final_value) + 'with a value of ' + str(accumulator[-1]) + 'after ' + str(self.num_iterations[
                          -1]) + 'iterations')

        best_value = np.argmin(accumulator)
        self.global_optimizer = final_values[best_value]
        if self.verbose >= 1:
            print('Descent algorithm performed, the optimal point found across ' + str(
                self.restarts) + 'different descents is ' + str(self.global_optimizer) + 'with a value of ' + str(
                np.min(accumulator)) + 'after ' + str(self.num_iterations[best_value]) + ' iterations')
        return self.trace_values


    def annealing(self, init_point):
        """ This function allows to produce points inside regions of importance in the optimization landscape of f.
        It enables to tune the initialized points entered by the user, or it allows to produce good points when the user
        did not input any initialized points. This allows to accelerate the optimization algorithm.
        For now, it only supports quadratic functions.
        INPUTS
        =======
        self: an instance of optimizer class
        init_point: the initial point, produced via self.produce_random_points() or input by user

        RETURNS
        ========
        accumulator[-1][-1]: list of values
           Has the form (x, y, z) where this point is the best point found accross the different epochs based
           on random sampling based on simulated annealing.

        EXAMPLES
        =========
        >>> f = lambda x: x**2+y**2+(z-2)**2
        >>> adam = Adam(f, random_restarts=10)
        >>> init_point = adam.produce_random_points()
        >>> init_point = adam.annealing(init_point)
        >>> adam.init = init_point
        >>> adam.descent()"""
        if self.covariance is None:
            raise ValueError('You must enter your quadratic form matrix in order to perform efficient sampling')

        temp = 10  # common default value when using this cooling schedule
        length = 500

        def reduce_temp(temp):
            return 0.8 * temp

        def incr_iters(length):
            return int(math.ceil(1.2 * length))

        old_solution = init_point
        old_cost = self.init_function(*old_solution)
        accepted = 0
        total = 0
        accumulator = []
        for epoch in range(self.max_epochs):
            if epoch > 0:
                temp = reduce_temp(temp)  # cooling schedule
                length = incr_iters(length)   # favor exploitation over exploration over time
            for it in range(length):
                total += 1
                new_solution = np.random.multivariate_normal(old_solution, self.covariance, size=1).flatten()
                new_cost = self.init_function(*new_solution)
                alpha = min(1, np.exp((old_cost - new_cost) / temp))
                if (new_cost < old_cost) or (np.random.uniform() < alpha):
                    accepted += 1
                    accumulator.append([temp, new_solution, new_cost])

                    old_cost = new_cost
                    old_solution = new_solution

                else:
                    accumulator.append([temp, old_solution, old_cost])
        if self.verbose == 2:
            if accepted * 1. / total < 0.1:
                print('The sampling has not yeat converged, you might want to increase the number of epochs')
            print('Initialization point before simulated annealing: '+str(self.init))
            print('Initialization point after simulated annealing: ' + str(accumulator[-1][1]))
        return accumulator[-1][1]


    def history(self):
        if self.global_optimizer is None:
            raise Exception('Optimization not run yet. Please run your optimizer before accessing to its history')
        return self.__str__()



class Adam(Optimizer):
    """
    This Adam class inherits the Optimizer class, leveraging the default values we defined for the parameters beta1 and
    beta2, which were, for the sake of simplicity, (since the users use mostly adam) the coefficients for the adam optimizer.
    EXAMPLES
    =========
    >>> f = lambda x, y, z: x ** 2 + y ** 2 + z ** 2
    >>> opt = Adam(f)
    """
    def __init__(self, f, max_iter=1e6, init_points=None, tolerance=1e-6, random_restarts=2, beta_1=0.9, beta_2=0.999,
                 step_size=1e-3, epsilon=1e-8, tuning=False, verbose=0):
        super(Adam, self).__init__(f, max_iter=max_iter, init_points=init_points, tolerance=1e-6, random_restarts=2, beta_1=0.9,
                                      beta_2=0.999,
                                      step_size=1e-3, epsilon=1e-8, tuning=False, verbose=0)
    def __str__(self):
        if self.global_optimizer:
            if self.vectorize:
                final_value = self.init_function(np.array(self.global_optimizer))
            else:
                final_value = self.init_function(*self.global_optimizer)
            return 'Adam optimizer, which found an optimal weight point of ' + str(
                self.global_optimizer) + ' the value of the objective function at this point is ' + str(
                final_value)
        else:
            return 'Adam optimizer, not yet fitted'


class sgd(Optimizer):
    """
    This sgd class inherits the Optimizer class, redefining the values for beta1 and beta2. For sgd, we do not want
    to leverage momentum, we just want to perform classic gradient descent, with a smaller batch size (addendum to be done later on).
    EXAMPLES
    =========
    >>> f = lambda x, y, z: x ** 2 + y ** 2 + z ** 2
    >>> opt = sgd(f)
    """

    def __init__(self, f, max_iter=1e6, init_points=None, tolerance=1e-6, random_restarts=2, beta_1=0.9, beta_2=0.999,
                 step_size=1e-3, epsilon=1e-8, tuning=False, verbose=0):
        super(sgd, self).__init__(f, max_iter=1e6, init_points=None, tolerance=1e-6, random_restarts=2, beta_1=0.9,
                                  beta_2=0.999,
                                  step_size=1e-3, epsilon=1e-8, tuning=False, verbose=0)
        self.beta1 = 0
        self.beta2 = 0


    def __str__(self):
        if self.global_optimizer:
            return 'Sgd optimizer, which found an optimal weight point of ' + str(
                self.global_optimizer) + ' the value of the objective function at this point is ' + str(
                self.init_function(*self.global_optimizer))
        else:
            return 'Sgd optimizer, not yet fitted'


class RMSProp(Optimizer):
    """
    This RMSprop class inherits the Optimizer class, redefining the values for beta1. For RMSProp, we do not
    want to put momentum on the gradient update, but which to have an adaptive learning rate depending on the
    accumulated squared gradients until then
     EXAMPLES
    =========
    >>> f = lambda x, y, z: x ** 2 + y ** 2 + z ** 2
    >>> opt = RMSProp(f)
    """

    def __init__(self, f, max_iter=1e6, init_points=None, tolerance=1e-6, random_restarts=2, beta_1=0.9, beta_2=0.999,
                 step_size=1e-3, epsilon=1e-8, tuning=False, verbose=0):
        super(RMSProp, self).__init__(f, max_iter=max_iter, init_points=init_points, tolerance=1e-6, random_restarts=2, beta_1=0.9,
                                      beta_2=0.999,
                                      step_size=1e-3, epsilon=1e-8, tuning=False, verbose=0)
        self.beta1 = 0

    def __str__(self):
        if self.global_optimizer:
            return 'RMSProp optimizer, which found an optimal weight point of ' + str(
                self.global_optimizer) + ' the value of the objective function at this point is ' + str(
                self.init_function(*self.global_optimizer))
        else:
            return 'RMSProp optimizer, not yet fitted'



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="CityAD", 
    version="0.0.16",
    author="Group CityScape",
    author_email="cityad.helpline@gmail.com",
    description="CityAD, an Automatic Differentiation package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cityscape-107/cs107-FinalProject",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy==1.18.0']
)